#include <util\atomic.h>			// need "--std=c99"
#include <avr\power.h>
#include <avr\interrupt.h>
#include <avr\sleep.h>
#include "humidity.h"
#include "getkey.h"


enum {
	WET,
	NORMAL,
	DRY,
	OFF
};

enum {
	FULLBATT,
	LOWBATT
};

uint8_t state = WET;
uint8_t batt = FULLBATT;


ISR( WAKEUP_vect )				// wake up from power down
{
  set_sleep_mode( SLEEP_MODE_IDLE );
  PCICR = 0;
}


static inline void power_down( void )
{
  ATOMIC_BLOCK(ATOMIC_FORCEON){
    if( get_key_idle( 1<<KEY_ON_OFF )){		// if debounce done
      ADCSRA = 0;				// ADC off
      PCICR = 1<<KEY_ON_OFF_ENABLE;		// enable wake up
      set_sleep_mode( SLEEP_MODE_PWR_DOWN );
    }
  }
}


static inline void init( void )
{

  PORTB = 0xFF;					// avoid floting pins, if unused
#ifdef _AVR_IOM48_H_
  PORTC = 0xFF;
  PORTD = 0xFF;
#else
  PORTA = 0xFF;
#endif
  KEY_ON_OFF_PCI_REG = 1<<KEY_ON_OFF_PCI;	// select wake up key
  set_sleep_mode( SLEEP_MODE_IDLE );		// default: idle
  sleep_enable();

  MEAS_PORT &= ~(1<<MEAS_IN);			// no pullup on analog input
  DIDR0 = 1<<MEAS_IN;				// analog input

  LED_WET_DDR = 1;
  LED_NORMAL_DDR = 1;
  LED_DRY_DDR = 1;
  LED_LOWBATT_DDR = 1;

  MEAS_PWR_DDR = 1;

  get_key_init();
  clock_prescale_set( clock_div_8 );		// 1MHz
}


static inline void power_on_off( void )		// handle key press
{
  if( get_key_press( 1<<KEY_ON_OFF )){
    if( state == OFF )
      state = WET;
    else
      state = OFF;
  }
}


uint16_t read_adc( uint8_t channel )
{
  uint16_t val = 0;

  ADMUX = channel;
  for( uint8_t i = 16; i; i-- ){
    ADCSRA = 1<<ADEN			// ADC on
	   | 1<<ADSC			// ADC start
	   | 1<<ADIF			// clear ADIF
	   | 1<<ADPS1 | 1<<ADPS0;	// 1MHz / 8 = 125kHz (50..200kHz)
    while( !(ADCSRA & 1<<ADIF) );	// wait until conversion done
    val += ADC;
  }
  return val >> 4;			// average over 16 samples
}


static inline void low_bat( void )
{
					// dummy measurement on Bandgap !!!
      	  read_adc( 1<<REFS0			// VCC as reference
	          | 0<<ADLAR			// ADC = 10 bit value
	          | 14 );			// input 14 = 1.1V Bandgap

  switch( read_adc( 1<<REFS0			// VCC as reference
	          | 0<<ADLAR			// ADC = 10 bit value
	          | 14 )){			// input 14 = 1.1V Bandgap

    case 0 ... THRESHOLD_LOWBATT1:
			batt = FULLBATT;
			break;

    case THRESHOLD_LOWBATT1 + 1 ... THRESHOLD_LOWBATT2:
			if( batt == FULLBATT )	// hysteresis
			  break;

    default:
			batt = LOWBATT;
  }
}


static inline void meas_humidity( void )
{
  MEAS_PWR = 1;					// sensor power on
  switch( read_adc( 1<<REFS0			// VCC as reference
		  | 0<<ADLAR			// ADC = 10 bit value
		  | MEAS_IN )){			// input ADC1

    case 0 ... THRESHOLD_WET1:
			state = WET;		// wet
			break;

    case THRESHOLD_WET1 + 1 ... THRESHOLD_WET2:
			if( state == WET )	// hysteresis (stay wet/normal)
			  break;

    case THRESHOLD_WET2 + 1 ... THRESHOLD_DRY1:
			state = NORMAL;		// normal
			break;

    case THRESHOLD_DRY1 + 1 ... THRESHOLD_DRY2:
			if( state == NORMAL )	// hysteresis (stay normal/dry)
			  break;

    default:
			state = DRY;		// dry
  }
  MEAS_PWR = 0;      				// sensor power off
}


int main( void )
{
  init();
  sei();

  for(;;){

    power_on_off();				// read on/off key
    meas_humidity();
    low_bat();

    if( batt != LOWBATT || flasher & 0xF0 )		// 4s
      LED_LOWBATT = LED_OFF;
    else
      LED_LOWBATT = LED_ON;

    switch( state ){

      case WET:
		if( flasher & 0x20 )			// 1s
		  LED_WET     = LED_OFF;
		else
		  LED_WET     = LED_ON;
		LED_NORMAL    = LED_OFF;
		LED_DRY       = LED_OFF;
		break;

      case NORMAL:
		LED_WET       = LED_OFF;
		LED_NORMAL    = LED_ON;
		LED_DRY       = LED_OFF;
		break;

      case DRY:
		LED_WET       = LED_OFF;
		LED_NORMAL    = LED_OFF;
		if( flasher & 0x10 )			// 0.5s
		  LED_DRY     = LED_OFF;
		else
		  LED_DRY     = LED_ON;
		break;

      case OFF:
		LED_WET       = LED_OFF;
		LED_NORMAL    = LED_OFF;
		LED_DRY       = LED_OFF;
		LED_LOWBATT   = LED_OFF;

		power_down();				// OFF = power down
		break;
    }
    sleep_cpu();
  }
}
